=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2007-02-12 05:09:45 +0000
+++ bzrlib/errors.py	2007-02-19 02:36:38 +0000
@@ -765,7 +765,7 @@
         self.obj = obj
 
 
-class TokenMismatch(LockError):
+class TokenMismatch(LockBroken):
 
     _fmt = "The lock token %(given_token)r does not match lock token %(lock_token)r."
 

=== modified file 'bzrlib/lockdir.py'
--- bzrlib/lockdir.py	2007-02-13 13:59:29 +0000
+++ bzrlib/lockdir.py	2007-02-19 02:36:38 +0000
@@ -444,6 +444,7 @@
             self.nonce = token
             self._lock_held = True
             self._locked_via_token = True
+            return token
         else:
             self.wait_lock()
             return self.peek().get('nonce')

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2007-02-13 13:59:29 +0000
+++ bzrlib/remote.py	2007-02-19 02:36:38 +0000
@@ -192,6 +192,7 @@
         self._lock_mode = None
         self._lock_token = None
         self._lock_count = 0
+        self._leave_lock = False
 
     def _ensure_real(self):
         """Ensure that there is a _real_repository set.
@@ -279,17 +280,34 @@
         if not self._lock_mode:
             self._lock_mode = 'r'
             self._lock_count = 1
+            if self._real_repository is not None:
+                self._real_repository.lock_read()
         else:
             self._lock_count += 1
-        self._ensure_real()
-        return self._real_repository.lock_read()
+
+    def _lock_write(self, token):
+        path = self.bzrdir._path_for_remote_call(self._client)
+        if token is None:
+            token = ''
+        response = self._client.call('Repository.lock_write', path, token)
+        if response[0] == 'ok':
+            ok, token = response
+            return token
+        elif response[0] == 'LockContention':
+            raise errors.LockContention('(remote lock)')
+        else:
+            assert False, 'unexpected response code %s' % (response,)
 
     def lock_write(self, token=None):
-        # definately wrong: want to check if there is a real repo
-        # and not thunk through if not
         if not self._lock_mode:
-            self._ensure_real()
-            self._lock_token = self._real_repository.lock_write(token=token)
+            self._lock_token = self._lock_write(token)
+            assert self._lock_token, 'Remote server did not return a token!'
+            if self._real_repository is not None:
+                self._real_repository.lock_write(token=self._lock_token)
+            if token is not None:
+                self._leave_lock = True
+            else:
+                self._leave_lock = False
             self._lock_mode = 'w'
             self._lock_count = 1
         elif self._lock_mode == 'r':
@@ -299,10 +317,10 @@
         return self._lock_token
 
     def leave_lock_in_place(self):
-        self._real_repository.leave_lock_in_place()
+        self._leave_lock = True
 
     def dont_leave_lock_in_place(self):
-        self._real_repository.dont_leave_lock_in_place()
+        self._leave_lock = False
 
     def _set_real_repository(self, repository):
         """Set the _real_repository for this repository.
@@ -316,17 +334,33 @@
             # acquire the lock with our token.
             self._real_repository.lock_write(self._lock_token)
 
+    def _unlock(self, token):
+        path = self.bzrdir._path_for_remote_call(self._client)
+        response = self._client.call('Repository.unlock', path, token)
+        if response == ('ok',):
+            return
+        elif response[0] == 'TokenMismatch':
+            raise errors.TokenMismatch(token, '(remote token)')
+        else:
+            assert False, 'unexpected response code %s' % (response,)
+
     def unlock(self):
-        # should free cache context.
         self._lock_count -= 1
         if not self._lock_count:
+            mode = self._lock_mode
             self._lock_mode = None
+            if self._real_repository is not None:
+                self._real_repository.unlock()
+            if mode != 'w':
+                return
+            assert self._lock_token, 'Locked, but no token!'
+            token = self._lock_token
             self._lock_token = None
-            return self._real_repository.unlock()
+            if not self._leave_lock:
+                self._unlock(token)
 
     def break_lock(self):
-        # should hand off to the network - or better yet, we should not
-        # allow stale network locks ?
+        # should hand off to the network
         self._ensure_real()
         return self._real_repository.break_lock()
 

=== modified file 'bzrlib/smart/repository.py'
--- bzrlib/smart/repository.py	2007-02-03 16:04:53 +0000
+++ bzrlib/smart/repository.py	2007-02-19 02:36:38 +0000
@@ -140,3 +140,33 @@
             return SmartServerResponse(('yes', ))
         else:
             return SmartServerResponse(('no', ))
+
+
+class SmartServerRepositoryLockWrite(SmartServerRepositoryRequest):
+
+    def do_repository_request(self, repository, token):
+        # XXX: this probably should not have a token.
+        if token == '':
+            token = None
+        try:
+            token = repository.lock_write(token=token)
+        except errors.LockContention, e:
+            return SmartServerResponse(('LockContention',))
+        repository.leave_lock_in_place()
+        repository.unlock()
+        if token is None:
+            token = ''
+        return SmartServerResponse(('ok', token))
+
+
+class SmartServerRepositoryUnlock(SmartServerRepositoryRequest):
+
+    def do_repository_request(self, repository, token):
+        try:
+            repository.lock_write(token=token)
+        except errors.TokenMismatch, e:
+            return SmartServerResponse(('TokenMismatch',))
+        repository.dont_leave_lock_in_place()
+        repository.unlock()
+        return SmartServerResponse(('ok',))
+

=== modified file 'bzrlib/smart/request.py'
--- bzrlib/smart/request.py	2007-02-15 23:03:02 +0000
+++ bzrlib/smart/request.py	2007-02-19 02:36:38 +0000
@@ -283,6 +283,10 @@
 request_handlers.register_lazy(
     'Repository.is_shared', 'bzrlib.smart.repository', 'SmartServerRepositoryIsShared')
 request_handlers.register_lazy(
+    'Repository.lock_write', 'bzrlib.smart.repository', 'SmartServerRepositoryLockWrite')
+request_handlers.register_lazy(
+    'Repository.unlock', 'bzrlib.smart.repository', 'SmartServerRepositoryUnlock')
+request_handlers.register_lazy(
     'rmdir', 'bzrlib.smart.vfs', 'RmdirRequest')
 request_handlers.register_lazy(
     'stat', 'bzrlib.smart.vfs', 'StatRequest')

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2007-02-03 10:11:04 +0000
+++ bzrlib/tests/__init__.py	2007-02-19 02:36:38 +0000
@@ -1249,6 +1249,17 @@
                     this_tree=wt_to)
         wt_to.add_parent_tree_id(branch_from.last_revision())
 
+    def reduceLockdirTimeout(self):
+        """Reduce the default lock timeout for the duration of the test, so that
+        if LockContention occurs during a test, it does so quickly.
+
+        Tests that expect to provoke LockContention errors should call this.
+        """
+        orig_timeout = bzrlib.lockdir._DEFAULT_TIMEOUT_SECONDS
+        def resetTimeout():
+            bzrlib.lockdir._DEFAULT_TIMEOUT_SECONDS = orig_timeout
+        self.addCleanup(resetTimeout)
+        bzrlib.lockdir._DEFAULT_TIMEOUT_SECONDS = 0
 
 BzrTestBase = TestCase
 

=== modified file 'bzrlib/tests/repository_implementations/test_repository.py'
--- bzrlib/tests/repository_implementations/test_repository.py	2007-02-15 01:32:11 +0000
+++ bzrlib/tests/repository_implementations/test_repository.py	2007-02-19 02:36:38 +0000
@@ -508,6 +508,12 @@
         repo.lock_write()
         repo.unlock()
 
+    def test_lock_read_then_unlock(self):
+        # Calling lock_read then unlocking should work without errors.
+        repo = self.make_repository('r')
+        repo.lock_read()
+        repo.unlock()
+
 
 class TestCaseWithComplexRepository(TestCaseWithRepository):
 

=== modified file 'bzrlib/tests/test_lockable_files.py'
--- bzrlib/tests/test_lockable_files.py	2007-02-13 13:59:29 +0000
+++ bzrlib/tests/test_lockable_files.py	2007-02-19 02:36:38 +0000
@@ -37,13 +37,7 @@
 class _TestLockableFiles_mixin(object):
 
     def setUp(self):
-        # Reduce the default timeout, so that if tests fail, they will do so
-        # reasonably quickly.
-        orig_timeout = lockdir._DEFAULT_TIMEOUT_SECONDS
-        def resetTimeout():
-            lockdir._DEFAULT_TIMEOUT_SECONDS = orig_timeout
-        self.addCleanup(resetTimeout)
-        lockdir._DEFAULT_TIMEOUT_SECONDS = 0
+        self.reduceLockdirTimeout()
 
     def test_read_write(self):
         self.assertRaises(NoSuchFile, self.lockable.get, 'foo')
@@ -145,6 +139,22 @@
         finally:
             self.lockable.unlock()
 
+    def test_lock_write_returns_token_when_given_token(self):
+        token = self.lockable.lock_write()
+        try:
+            if token is None:
+                # This test does not apply, because this lockable refuses
+                # tokens.
+                return
+            new_lockable = self.get_lockable()
+            token_from_new_lockable = new_lockable.lock_write(token=token)
+            try:
+                self.assertEqual(token, token_from_new_lockable)
+            finally:
+                new_lockable.unlock()
+        finally:
+            self.lockable.unlock()
+
     def test_lock_write_raises_on_token_mismatch(self):
         token = self.lockable.lock_write()
         try:
@@ -388,3 +398,7 @@
     def test_dont_leave_in_place(self):
         # See test_lock_write_returns_None_refuses_token.
         return None
+
+    def test_lock_write_returns_token_when_given_token(self):
+        # See test_lock_write_returns_None_refuses_token.
+        return None

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2007-02-08 06:51:30 +0000
+++ bzrlib/tests/test_remote.py	2007-02-19 02:36:38 +0000
@@ -23,10 +23,14 @@
 
 from cStringIO import StringIO
 
-from bzrlib import bzrdir, remote, tests
+from bzrlib import (
+    bzrdir,
+    errors,
+    remote,
+    tests,
+    )
 from bzrlib.branch import Branch
 from bzrlib.bzrdir import BzrDir, BzrDirFormat
-from bzrlib.errors import NoSuchRevision
 from bzrlib.remote import (
     RemoteBranch,
     RemoteBzrDir,
@@ -202,7 +206,7 @@
         branch = RemoteBranch(bzrdir, None, _client=client)
 
         self.assertRaises(
-            NoSuchRevision, branch.set_revision_history, ['rev-id'])
+            errors.NoSuchRevision, branch.set_revision_history, ['rev-id'])
 
 
 class TestBranchControlGetBranchConf(tests.TestCase):
@@ -359,7 +363,7 @@
         repo, client = self.setup_fake_client_and_repository(
             responses, transport_path)
         # also check that the right revision is reported in the error
-        self.assertRaises(NoSuchRevision,
+        self.assertRaises(errors.NoSuchRevision,
             repo.get_revision_graph, revid)
         self.assertEqual(
             [('call2', 'Repository.get_revision_graph', ('///sinhala/', revid))],
@@ -391,3 +395,55 @@
             [('call', 'Repository.is_shared', ('///qwack/',))],
             client._calls)
         self.assertEqual(False, result)
+
+
+class TestRepositoryLockWrite(TestRemoteRepository):
+
+    def test_lock_write(self):
+        responses = [(('ok', 'a token'), '')]
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(
+            responses, transport_path)
+        result = repo.lock_write()
+        self.assertEqual(
+            [('call', 'Repository.lock_write', ('///quack/',))],
+            client._calls)
+        self.assertEqual('a token', result)
+
+    def test_lock_write_already_locked(self):
+        responses = [(('LockContention', ), '')]
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(
+            responses, transport_path)
+        self.assertRaises(errors.LockContention, repo.lock_write)
+        self.assertEqual(
+            [('call', 'Repository.lock_write', ('///quack/',))],
+            client._calls)
+
+
+class TestRepositoryUnlock(TestRemoteRepository):
+
+    def test_unlock(self):
+        responses = [(('ok', 'a token'), ''),
+                     (('ok',), '')]
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(
+            responses, transport_path)
+        repo.lock_write()
+        repo.unlock()
+        self.assertEqual(
+            [('call', 'Repository.lock_write', ('///quack/',)),
+             ('call', 'Repository.unlock', ('///quack/', 'a token'))],
+            client._calls)
+
+    def test_unlock_wrong_token(self):
+        # If somehow the token is wrong, unlock will raise TokenMismatch.
+        responses = [(('ok', 'a token'), ''),
+                     (('TokenMismatch',), '')]
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(
+            responses, transport_path)
+        repo.lock_write()
+        self.assertRaises(errors.TokenMismatch, repo.unlock)
+
+

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2007-02-15 23:03:02 +0000
+++ bzrlib/tests/test_smart.py	2007-02-19 02:36:38 +0000
@@ -467,6 +467,67 @@
             request.execute(backing.local_abspath(''), ))
 
 
+class TestSmartServerRepositoryLockWrite(tests.TestCaseWithTransport):
+
+    def setUp(self):
+        tests.TestCaseWithTransport.setUp(self)
+        self.reduceLockdirTimeout()
+
+    def test_lock_write_on_unlocked_repo(self):
+        backing = self.get_transport()
+        request = smart.repository.SmartServerRepositoryLockWrite(backing)
+        repository = self.make_repository('.')
+        response = request.execute(backing.local_abspath(''))
+        nonce = repository.control_files._lock.peek().get('nonce')
+        self.assertEqual(SmartServerResponse(('ok', nonce)), response)
+        # The repository is now locked.  Verify that with a new repository
+        # object.
+        new_repo = repository.bzrdir.open_repository()
+        self.assertRaises(errors.LockContention, new_repo.lock_write)
+
+    def test_lock_write_on_locked_repo(self):
+        backing = self.get_transport()
+        request = smart.repository.SmartServerRepositoryLockWrite(backing)
+        repository = self.make_repository('.')
+        repository.lock_write()
+        repository.leave_lock_in_place()
+        repository.unlock()
+        response = request.execute(backing.local_abspath(''))
+        self.assertEqual(
+            SmartServerResponse(('LockContention',)), response)
+
+
+class TestSmartServerRepositoryUnlock(tests.TestCaseWithTransport):
+
+    def setUp(self):
+        tests.TestCaseWithTransport.setUp(self)
+        self.reduceLockdirTimeout()
+
+    def test_unlock_on_locked_repo(self):
+        backing = self.get_transport()
+        request = smart.repository.SmartServerRepositoryUnlock(backing)
+        repository = self.make_repository('.')
+        token = repository.lock_write()
+        repository.leave_lock_in_place()
+        repository.unlock()
+        response = request.execute(backing.local_abspath(''), token)
+        self.assertEqual(
+            SmartServerResponse(('ok',)), response)
+        # The repository is now unlocked.  Verify that with a new repository
+        # object.
+        new_repo = repository.bzrdir.open_repository()
+        new_repo.lock_write()
+        new_repo.unlock()
+
+    def test_unlock_on_unlocked_repo(self):
+        backing = self.get_transport()
+        request = smart.repository.SmartServerRepositoryUnlock(backing)
+        repository = self.make_repository('.')
+        response = request.execute(backing.local_abspath(''), 'some token')
+        self.assertEqual(
+            SmartServerResponse(('TokenMismatch',)), response)
+
+
 class TestHandlers(tests.TestCase):
     """Tests for the request.request_handlers object."""
 
@@ -505,3 +566,9 @@
         self.assertEqual(
             smart.request.request_handlers.get('Repository.is_shared'),
             smart.repository.SmartServerRepositoryIsShared)
+        self.assertEqual(
+            smart.request.request_handlers.get('Repository.lock_write'),
+            smart.repository.SmartServerRepositoryLockWrite)
+        self.assertEqual(
+            smart.request.request_handlers.get('Repository.unlock'),
+            smart.repository.SmartServerRepositoryUnlock)

